"""
Compute expression counts

This command takes a table of filtered IgBLAST results (filtered.tab.gz),
filters it by various criteria and then counts how often specific
genes are named.

By default, all genes with non-zero expression are listed, sorted by
name. Use --database to change this.
"""
import sys
import logging
import pandas as pd
from sqt import SequenceReader
from .table import read_table
from .utils import natural_sort_key
from .discoverj import AlleleRatioMerger

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--gene', choices=('V', 'D', 'J'), default='V',
		help='Which gene type: Choose V, D or J. Default: Default: %(default)s')
	arg('--database', metavar='FASTA',
		help='Compute expressions for the sequences that are named in the FASTA '
		'file. Only the sequence names in the file are used! This is the only '
		'way to also include genes with an expression of zero.')
	arg('--plot', metavar='FILE',
		help='Plot expressions to FILE (PDF or PNG)')

	group = parser.add_argument_group('Input table filtering')
	group.add_argument('--d-evalue', type=float, default=None,
		help='Maximal allowed E-value for D gene match. Default: 1E-4 '
		'if --gene=D, no restriction otherwise.')
	group.add_argument('--d-coverage', '--D-coverage', type=float, default=None,
		help='Minimum D coverage (in percent). Default: 70 '
		'if --gene=D, no restriction otherwise.')
	group.add_argument('--d-errors', type=int, default=None,
		help='Maximum allowed D errors. Default: No limit.')

	group = parser.add_argument_group('Expression counts table filtering')
	group.add_argument('--allele-ratio', type=float, metavar='RATIO', default=None,
		help='Required allele ratio. Works only for genes named "NAME*ALLELE". '
		'Default: Do not check allele ratio.')

	arg('table', help='Table with parsed and filtered IgBLAST results')


def count_unique_cdr3(table):
	return len(set(s for s in table.CDR3_nt if s))


def count_unique_gene(table, gene_type):
	return len(set(s for s in table[gene_type + '_gene'] if s))


def compute_expressions(table, gene_type):
	columns = ('gene', 'count', 'unique_CDR3')
	for gt in 'V', 'D', 'J':
		if gene_type != gt:
			columns += ('unique_' + gt,)
	gene_column = gene_type + '_gene'
	rows = []
	for gene, group in table.groupby(gene_column):
		if gene == '':
			continue
		unique_cdr3 = count_unique_cdr3(group)
		row = dict(gene=gene, count=len(group), unique_CDR3=unique_cdr3)
		for gt in 'V', 'D', 'J':
			if gene_type != gt:
				row['unique_' + gt] = count_unique_gene(group, gene_type=gt)
		rows.append(row)
	counts = pd.DataFrame(rows, columns=columns).set_index('gene')
	return counts


def plot_counts(counts, gene_type):
	"""Plot expression counts. Return a Figure object"""
	import matplotlib
	matplotlib.use('agg')
	import matplotlib.pyplot as plt
	import seaborn as sns
	import numpy as np

	fig = plt.figure(figsize=((50 + len(counts) * 5) / 25.4, 210/25.4))
	matplotlib.rcParams.update({'font.size': 14})
	ax = fig.gca()
	ax.set_title('{} gene usage'.format(gene_type))
	ax.set_xlabel('{} gene'.format(gene_type))
	ax.set_ylabel('Count')
	ax.set_xticks(np.arange(len(counts)) + 0.5)
	ax.set_xticklabels(counts.index, rotation='vertical')
	ax.grid(axis='x')
	ax.set_xlim((-0.25, len(counts)))
	ax.bar(np.arange(len(counts)), counts['count'])
	fig.set_tight_layout(True)
	return fig


def main(args):
	if args.database:
		with SequenceReader(args.database) as fr:
			gene_names = [record.name for record in fr]
		gene_names.sort(key=natural_sort_key)
	else:
		gene_names = None

	usecols = ['V_gene', 'D_gene', 'J_gene', 'V_errors', 'D_errors', 'J_errors', 'D_covered',
		'D_evalue', 'CDR3_nt']
	table = read_table(args.table, usecols=usecols)
	logger.info('Table with %s rows read', len(table))

	# Set default filters depending on gene
	if args.gene == 'D':
		if args.d_evalue is None:
			args.d_evalue = 1E-4
		if args.d_coverage is None:
			args.d_coverage = 70
	if args.d_evalue is not None:
		table = table[table.D_evalue <= args.d_evalue]
		logger.info('%s rows remain after requiring D E-value <= %s', len(table), args.d_evalue)

	if args.d_coverage is not None:
		table = table[table.D_covered >= args.d_coverage]
		logger.info('%s rows remain after requiring D coverage >= %s', len(table), args.d_coverage)

	if args.d_errors is not None:
		table = table[table.D_errors <= args.d_errors]
		logger.info('%s rows remain after requiring D errors <= %s', len(table), args.d_errors)

	logger.info('Computing expression counts for %s genes', args.gene)
	counts = compute_expressions(table, args.gene)

	# Make sure that always all gene names are listed even if no sequences
	# were assigned.
	if gene_names:
		counts = counts.reindex(gene_names, fill_value=0)

	if args.database and args.allele_ratio:
		logger.error('--database and --allele-ratio cannot be used at the same time.')
		sys.exit(1)

	if args.allele_ratio is not None:
		arm = AlleleRatioMerger(args.allele_ratio, cross_mapping_ratio=None)
		renamed_counts = counts.reset_index().rename(columns={'gene': 'name'})
		arm.extend(renamed_counts.itertuples(index=False))
		counts = pd.DataFrame(list(arm), columns=renamed_counts.columns) \
			.rename(columns={'name': 'gene'}) \
			.set_index('gene')
		logger.info(
			'After filtering by allele ratio and/or cross-mapping ratio, %d candidates remain',
			len(counts))

	# logger.info('%d sequences were not assigned a %s gene', unassigned, args.gene)
	counts.to_csv(sys.stdout, sep='\t')
	logger.info('Wrote expression count table')

	if args.plot:
		plot_counts(counts, args.gene).savefig(args.plot)
		logger.info("Wrote %s", args.plot)



"""
Some utility functions that work on sequences and lists of sequences.
"""
import os
import sys
import re
import random
import hashlib
import resource
from collections import OrderedDict

import numpy as np
from sqt.align import edit_distance, multialign, consensus
from sqt.dna import GENETIC_CODE, nt_to_aa as _nt_to_aa
from sqt import SequenceReader
from cutadapt.align import Aligner


def downsampled(population, size):
	"""
	Return a random subsample of the population.

	Uses reservoir sampling. See https://en.wikipedia.org/wiki/Reservoir_sampling
	"""
	sample = population[:size]
	for index in range(size, len(population)):
		r = random.randint(0, index)
		if r < size:
			sample[r] = population[index]
	return sample


def distances(sequences, band=0.2):
	"""
	Compute all pairwise edit distances and return a square matrix.

	Entry [i,j] in the matrix is the edit distance between sequences[i]
	and sequences[j].
	"""
	m = np.zeros((len(sequences), len(sequences)), dtype=float)
	maxdiff = max((int(len(s) * band) for s in sequences), default=0)
	for i, s in enumerate(sequences):
		for j, t in enumerate(sequences):
			if i < j:
				m[j, i] = m[i, j] = min(maxdiff+1, edit_distance(s, t, maxdiff=maxdiff))
	return m


def iterative_consensus(sequences, program='muscle-medium', threshold=0.6,
		subsample_size=200, maximum_subsample_size=1600):
	"""
	Compute a consensus sequence of the given sequences, but do not use all
	sequences if there are many: First, try to compute the consensus from a
	small subsample. If there are 'N' bases, increase the subsample size and
	repeat until either there are no more 'N' bases, all available sequences
	have been used or maximum_subsample_size is reached.
	"""
	while True:
		sample = downsampled(sequences, subsample_size)
		aligned = multialign(OrderedDict(enumerate(sample)), program=program)
		cons = consensus(aligned, threshold=threshold).strip('N')
		if 'N' not in cons:
			# This consensus is good enough
			break
		if len(sequences) <= subsample_size:
			# We have already used all the sequences that are available
			break
		subsample_size *= 2
		if subsample_size > maximum_subsample_size:
			break
	return cons


def sequence_hash(s, digits=4):
	"""
	For a string, return a 'fingerprint' that looks like 'S1234' (the character
	'S' is fixed). The idea is that this allows one to quickly see whether two
	sequences are not identical.
	"""
	h = int(hashlib.md5(s.encode()).hexdigest()[-4:], base=16)
	return 'S' + str(h % 10**digits).rjust(digits, '0')


def unique_name(name, sequence):
	"""Create a unique name based on the current name and the sequence

	The returned name looks like name_S1234. If the current name contains
	already a S_.... suffix, it is removed before the new suffix is appended.

	name -- current name
	"""
	return '{}_{}'.format(name.rsplit('_S', 1)[0], sequence_hash(sequence))


class SerialPool:
	"""
	An alternative to multiprocessing.Pool that runs things in serial for
	easier debugging
	"""
	def __init__(self, *args, **kwargs):
		pass

	def __enter__(self):
		return self

	def __exit__(self, *args):
		pass

	def imap(self, func, iterable, chunksize):
		for i in iterable:
			yield func(i)


def natural_sort_key(s, _nsre=re.compile('([0-9]+)')):
	"""
	Use this function as sorting key to sort in 'natural' order.

	>>> names = ['file10', 'file1.5', 'file1']
	>>> sorted(names, key=natural_sort_key)
	['file1', 'file1.5', 'file10']

	Source: http://stackoverflow.com/a/16090640/715090
	"""
	return [int(text) if text.isdigit() else text.lower()
		for text in re.split(_nsre, s)]


class UniqueNamer:
	"""
	Assign unique names by appending letters to already seen names.
	"""
	def __init__(self):
		self._names = set()

	def __call__(self, name):
		ext = 'A'
		new_name = name
		while new_name in self._names:
			if ext == '[':
				raise ValueError('Too many duplicate names')
			new_name = name + ext
			ext = chr(ord(ext) + 1)
		self._names.add(new_name)
		return new_name


class Merger:
	"""
	Merge similar items into one. To specify what "similar" means, implement
	the merged() method in a subclass.
	"""
	def __init__(self):
		self._items = []

	def add(self, item):
		# This method could possibly be made simpler if the graph structure
		# was made explicit.

		items = []
		for existing_item in self._items:
			m = self.merged(existing_item, item)
			if m is None:
				items.append(existing_item)
			else:
				item = m
		items.append(item)
		self._items = items

	def extend(self, iterable):
		for i in iterable:
			self.add(i)

	def __iter__(self):
		if self._items and hasattr(self._items, 'name'):
			yield from sorted(self._items, key=lambda x: x.name)
		else:
			yield from self._items

	def __len__(self):
		return len(self._items)

	def merged(self, existing_item, item):
		"""
		If existing_item and item can be returned, this method must return
		a new item that represents a merged version of both. If they cannot
		be merged, it must return None.
		"""
		raise NotImplementedError("not implemented")


def relative_symlink(src, dst, force=False):
	"""
	Create a symbolic link in any directory.

	force -- if True, then overwrite an existing file/symlink
	"""
	if force:
		try:
			os.remove(dst)
		except FileNotFoundError:
			pass
	target = os.path.relpath(os.path.abspath(src), start=os.path.dirname(dst))
	os.symlink(target, dst)


def nt_to_aa(s):
	"""Translate a nucleotide sequence to an amino acid sequence"""
	try:
		# try fast version first
		return _nt_to_aa(s)
	except ValueError:
		# failure because there was an unknown nucleotide
		return ''.join(GENETIC_CODE.get(s[i:i+3], '*') for i in range(0, len(s), 3))


def has_stop(sequence):
	"""
	Return a boolean indicating whether the sequence has an internal stop codon.
	An incomplete codon at the end is allowed.

	>>> has_stop('GGG')
	False
	>>> has_stop('TAA')
	True
	>>> has_stop('GGGAC')
	False
	"""
	s = sequence[:len(sequence) // 3 * 3]
	return '*' in nt_to_aa(s)


def plural_s(n):
	return 's' if n != 1 else ''


class FastaValidationError(Exception):
	pass


def validate_fasta(path):
	"""
	Ensure that the FASTA file is suitable for use with makeblastdb.
	Raise a FastaValidationError if any of the following are true:

	- a record is empty
	- a record name occurs more than once
	- a sequence occurs more than once
	"""
	with SequenceReader(path) as sr:
		records = list(sr)

	names = set()
	sequences = dict()
	for r in records:
		if len(r.sequence) == 0:
			raise FastaValidationError("Record {!r} is empty".format(r.name))
		if r.name in names:
			raise FastaValidationError("Record name {!r} occurs more than once".format(r.name))
		s = r.sequence.upper()
		if s in sequences:
			raise FastaValidationError("Records {!r} and {!r} contain the same sequence".format(
				r.name, sequences[s]))
		sequences[s] = r.name
		names.add(r.name)


def find_overlap(s, t, min_overlap=1):
	"""
	Detect if s and t overlap.

	Returns:

	None if no overlap was detected.
	0 if s is a prefix of t or t is a prefix of s.
	Positive int gives index where t starts within s.
	Negative int gives -index where s starts within t.

	>>> find_overlap('ABCDE', 'CDE')
	2
	>>> find_overlap('CDE', 'ABCDEFG')
	-2
	>>> find_overlap('ABC', 'X') is None
	True
	"""
	aligner = Aligner(s, max_error_rate=0)
	aligner.min_overlap = min_overlap
	result = aligner.locate(t)
	if result is None:
		return None
	s_start, _, t_start, _, _, _ = result
	return s_start - t_start


def merge_overlapping(s, t):
	"""
	Return merged sequences or None if they do not overlap
	The minimum overlap is 50% of the length of the shorter sequence.
	"""
	i = find_overlap(s, t, min_overlap=max(1, min(len(s), len(t)) // 2))
	if i is None:
		return None

	if i >= 0:
		# positive: index of t in s
		if i + len(t) < len(s):
			# t is in s
			return s
		return s[:i] + t
	if -i + len(s) < len(t):
		# s is in t
		return t
	return t[:-i] + s


def get_cpu_time():
	"""Return CPU time used by process and children"""
	if sys.platform != 'linux':
		return None
	rs = resource.getrusage(resource.RUSAGE_SELF)
	rc = resource.getrusage(resource.RUSAGE_CHILDREN)
	return rs.ru_utime + rs.ru_stime + rc.ru_utime + rc.ru_stime


def slice_arg(s):
	"""
	Parse a string that describes a slice with start and end.

	>>> slice_arg('2:-3')
	slice(2, -3, None)

	>> slice_arg(':-3')
	slice(None, -3, None)

	>> slice_arg('2:')
	slice(2, None, None)
	"""

	start, end = s.split(':')
	start = None if start == '' else int(start)
	end = None if end == '' else int(end)
	return slice(start, end)


def is_same_gene(name1: str, name2: str):
	"""
	Compare gene names to find out whether they are alleles of each other.
	Both names must have a '*' in them
	"""
	return '*' in name1 and '*' in name2 and name1.split('*')[0] == name2.split('*')[0]



"""
Function for reading the table created by the 'parse' subcommand.
"""
import logging
import pandas as pd

logger = logging.getLogger(__name__)


# These columns contain string data
# convert them to str to avoid a PerformanceWarning
# TODO some of these are actually categorical or bool
_STRING_COLUMNS = [
	'V_gene',  # categorical
	'D_gene',  # categorical
	'J_gene',  # categorical
	'chain',  # categorical
	'stop',  # bool
	'productive',  # bool
	'UTR',
	'leader',
	'CDR1_nt',
	'CDR1_aa',
	'CDR2_nt',
	'CDR2_aa',
	'CDR3_nt',
	'CDR3_aa',
	'V_nt',
	'V_aa',
	'V_end',
	'VD_junction',
	'D_region',
	'DJ_junction',
	'J_nt',
	'VDJ_nt',
	'VDJ_aa',
	'name',
	'barcode',
	'race_G',
	'genomic_sequence',
]

_INTEGER_COLUMNS = ('V_errors', 'D_errors', 'J_errors', 'V_CDR3_start')


def fix_columns(df):
	"""
	Changes DataFrame in-place
	"""
	# Convert all string columns to str to avoid a PerformanceWarning
	for col in _STRING_COLUMNS:
		if col not in df:
			continue
		df[col].fillna('', inplace=True)
		df[col] = df[col].astype('str')
		# Empty strings have been set to NaN by read_csv. Replacing
		# by the empty string avoids problems with groupby, which
		# ignores NaN values.
	# Columns that have any NaN values in them cannot be converted to
	# int due to a numpy limitation.
	for col in _INTEGER_COLUMNS:
		if col not in df.columns:
			continue
		if all(df[col].notnull()):
			df[col] = df[col].astype(int)

	# TODO backwards compatibility
	if 'CDR3_clusters' in df.columns:
		df.rename(columns={'CDR3_clusters': 'clonotypes'}, inplace=True)


def read_table(path, usecols=None, log=False):
	"""
	Read in the table created by the parse subcommand (typically named *.tab)
	"""
	d = pd.read_csv(path, usecols=usecols, sep='\t')
	fix_columns(d)
	if log:
		logger.info('%s rows in input table', len(d))
	return d



from collections import OrderedDict, defaultdict
import pandas as pd
from scipy.spatial import distance
from scipy.cluster import hierarchy
from sqt.align import hamming_distance
from .utils import distances
from .trie import Trie


def all_nodes(root):
	"""Return a list of all nodes of the tree, from left to right (iterative implementation)."""
	result = []
	path = [None]
	node = root
	while node is not None:
		if node.left is not None:
			path.append(node)
			node = node.left
		elif node.right is not None:
			result.append(node)
			node = node.right
		else:
			result.append(node)
			node = path.pop()
			if node is not None:
				result.append(node)
				node = node.right

	return result


def inner_nodes(root):
	"""
	Return a list of all inner nodes of the tree, from left to right.
	"""
	return [node for node in all_nodes(root) if not node.is_leaf()]


def collect_ids(root):
	"""
	Return a list of ids of all leaves of the given tree
	"""
	return [node.id for node in all_nodes(root) if node.is_leaf()]


def cluster_sequences(sequences, minsize=5):
	"""
	Cluster the given sequences into groups of similar sequences.

	Return a triple that contains a pandas.DataFrame with the edit distances,
	the linkage result, and a list that maps sequence ids to their cluster id.
	If an entry is zero in that list, it means that the sequence is not part of
	a cluster.
	"""
	matrix = distances(sequences)
	linkage = hierarchy.linkage(distance.squareform(matrix), method='average')
	# Linkage columns are:
	# 0, 1: merged clusters, 2: distance, 3: number of nodes in cluster
	inner = inner_nodes(hierarchy.to_tree(linkage))
	prev = linkage[:, 2].max()  # highest distance
	clusters = [0] * len(sequences)
	cl = 1
	for n in inner:
		if n.dist > 0 and prev / n.dist < 0.8 \
				and n.left.count >= minsize and n.right.count >= minsize:
			for id in collect_ids(n.left):
				# Do not overwrite previously assigned ids
				if clusters[id] == 0:
					clusters[id] = cl
			cl += 1
		prev = n.dist
	# At the end of the above loop, we have not processed the rightmost
	# subtree. In our experiments, it never contains true novel sequences,
	# so we omit it.

	return pd.DataFrame(matrix), linkage, clusters


class Graph:
	"""Graph that can find connected components"""
	def __init__(self, nodes):
		self._nodes = OrderedDict()
		for node in nodes:
			self._nodes[node] = []

	def add_edge(self, node1, node2):
		self._nodes[node1].append(node2)
		self._nodes[node2].append(node1)

	def connected_components(self):
		"""Return a list of connected components."""
		visited = set()
		components = []
		for node, neighbors in self._nodes.items():
			if node in visited:
				continue
			# Start a new component
			to_visit = [node]
			component = []
			while to_visit:
				n = to_visit.pop()
				if n in visited:
					continue
				visited.add(n)
				component.append(n)
				for neighbor in self._nodes[n]:
					if neighbor not in visited:
						to_visit.append(neighbor)
			components.append(component)
		return components


def cluster_by_length(strings):
	"""
	Cluster a set of strings by length
	"""
	string_lists = defaultdict(list)
	for s in strings:
		string_lists[len(s)].append(s)
	return list(string_lists.values())


def single_linkage(strings, linked):
	"""
	Cluster a set of strings. *linked* is a function with two parameters s and t that returns
	whether *s* and *t* are in the same cluster.

	>>> single_linkage(['ABC', 'ABD', 'DEFG', 'DEFH'], lambda s, t: len(s) == len(t) and
	...		hamming_distance(s, t) <= 1)
	[['ABC', 'ABD'], ['DEFG', 'DEFH']]
	"""
	graph = Graph(strings)
	for i, s in enumerate(strings):
		for j, t in enumerate(strings[i + 1:]):
			if linked(s, t):
				graph.add_edge(s, t)
	return graph.connected_components()


def hamming_single_linkage(strings, mismatches, linked=None):
	"""
	Cluster a set of strings by hamming distance. Strings with
	a distance of at most 'mismatches' will be put into the same cluster.

	Uses the optimization that strings of different lengths can be
	clustered separately.

	Use *linked* to override the function passed to single_linkage. It will
	only be called for strings of the same length.

	Return a list of connected components (clusters).
	"""
	if linked is None:
		def linked(s, t):
			return hamming_distance(s, t) <= mismatches

	components = []
	for strings in cluster_by_length(strings):
		components.extend(single_linkage(strings, linked))
	return components


def hamming_single_linkage_trie(strings, mismatches):
	"""
	Cluster by hamming distance using a trie
	"""
	components = []
	for strings in cluster_by_length(strings):
		graph = Graph(strings)
		trie = Trie()
		for s in strings:
			trie.add(s)
		for s in strings:
			for neighbor in trie.find_all_similar(s, mismatches):
				if neighbor != s:
					graph.add_edge(s, neighbor)
		components.extend(graph.connected_components())
	return components



"""
Plot histograms of differences to reference V gene

For each gene, a histogram is plotted that shows how often a sequence was
assigned to that gene at a certain percentage difference.
"""
import logging
import matplotlib
matplotlib.use('agg')
from matplotlib.backends.backend_pdf import FigureCanvasPdf, PdfPages
from matplotlib.figure import Figure
import numpy as np
from .table import read_table

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--minimum-group-size', '-m', metavar='N', default=None, type=int,
		help="Plot only genes with at least N assigned sequences. "
		"Default: 0.1% of assigned sequences or 100, whichever is smaller.")
	arg('--ignore-J', action='store_true', default=False,
		help='Include also rows without J assignment or J%%SHM>0.')
	arg('table', help='Table with parsed IgBLAST results')
	arg('pdf', help='Plot error frequency histograms to this PDF file', default=None)


def plot_difference_histogram(group, gene_name, bins=np.arange(20.1)):
	"""
	Plot a histogram of percentage differences for a specific gene.
	"""
	exact_matches = group[group.V_SHM == 0]
	CDR3s_exact = len(set(s for s in exact_matches.CDR3_nt if s))
	Js_exact = len(set(exact_matches.J_gene))

	fig = Figure(figsize=(100/25.4, 60/25.4))
	ax = fig.gca()
	ax.set_xlabel('Percentage difference')
	ax.set_ylabel('Frequency')
	fig.suptitle('Gene ' + gene_name, y=1.08, fontsize=16)
	ax.set_title('{:,} sequences assigned'.format(len(group)))

	ax.text(0.25, 0.95,
		'{:,} ({:.1%}) exact matches\n  {} unique CDR3\n  {} unique J'.format(
			len(exact_matches), len(exact_matches) / len(group),
			CDR3s_exact, Js_exact),
		transform=ax.transAxes, fontsize=10,
		bbox=dict(boxstyle='round', facecolor='white', alpha=0.5),
		horizontalalignment='left', verticalalignment='top')
	_ = ax.hist(list(group.V_SHM), bins=bins)
	return fig


def main(args):
	table = read_table(args.table)

	# Discard rows with any mutation within J at all
	logger.info('%s rows read', len(table))
	if not args.ignore_J:
		# Discard rows with any mutation within J at all
		table = table[table.J_SHM == 0][:]
		logger.info('%s rows remain after discarding J%%SHM > 0', len(table))

	if args.minimum_group_size is None:
		total = len(table)
		minimum_group_size = min(total // 1000, 100)
		logger.info('Skipping genes with less than %s assignments', minimum_group_size)
	else:
		minimum_group_size = args.minimum_group_size
	n = 0
	too_few = 0
	with PdfPages(args.pdf) as pages:
		for gene, group in table.groupby('V_gene'):
			if len(group) < minimum_group_size:
				too_few += 1
				continue
			fig = plot_difference_histogram(group, gene)
			n += 1
			FigureCanvasPdf(fig).print_figure(pages, bbox_inches='tight')
	logger.info('%s plots created (%s skipped because of too few sequences)', n, too_few)



"""
Filter V gene candidates (germline and pre-germline filter)

After candidates for novel V genes have been found with the 'discover'
subcommand, this script is used to filter the candidates and make sure that
only true germline genes remain ("germline filter" and "pre-germline filter").
The following filtering and processing steps are performed on each candidate
separately:

* Discard sequences with N bases
* Discard sequences that come from a consensus over too few source sequences (unless whitelisted)
* Discard sequences with too few unique CDR3s (CDR3_clusters column)
* Discard sequences with too few unique Js (Js_exact column)
* Discard sequences identical to one of the database sequences (if DB given)
* Discard sequences that do not match a set of known good motifs (unless whitelisted)
* Discard sequences that contain a stop codon (has_stop column) (unless whitelisted)

The following criteria involve comparison of candidates against each other:

* Discard sequences that are too similar to another (unless whitelisted)
* Discard sequences that are cross-mapping artifacts
* Discard sequences that have a too low allele ratio
* Discard sequences with too few unique Ds relative to other alleles (Ds_exact column)

If you provide a whitelist of sequences, then the candidates that appear on it
* are not checked for the cluster size criterion,
* do not need to match a set of known good motifs,
* are never considered near-duplicates,
* are allowed to contain a stop codon.

The filtered table is written to standard output.
"""
import sys
import logging
from collections import namedtuple
import pandas as pd
from sqt import FastaReader
from sqt.align import edit_distance

from .utils import UniqueNamer, Merger, is_same_gene

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--cluster-size', type=int, metavar='N', default=0,
		help='Consensus must represent at least N sequences. '
		'Default: %(default)s')
	arg('--max-differences', type=int, metavar='MAXDIFF', default=1,
		help='Merge sequences if they have at most MAXDIFF differences. '
		'The one with more CDR3s is kept. Default: %(default)s')
	arg('--cross-mapping-ratio', type=float, metavar='RATIO', default=0.02,
		help='Ratio for detection of cross-mapping artifacts. Default: %(default)s')
	arg('--clonotype-ratio', '--allele-ratio', type=float, metavar='RATIO', default=0.1,
		help='Required ratio of "clonotypes" counts between alleles. '
		     'Works only for genes named "NAME*ALLELE". Default: %(default)s')
	arg('--exact-ratio', type=float, metavar='RATIO', default=0.1,
		help='Required ratio of "exact" counts between alleles. '
		     'Works only for genes named "NAME*ALLELE". Default: %(default)s')
	arg('--minimum-db-diff', '-b', type=int, metavar='N', default=0,
		help='Sequences must have at least N differences to the database '
		'sequence. Default: %(default)s')
	arg('--maximum-N', '-N', type=int, metavar='COUNT', default=0,
		help='Sequences must have at most COUNT "N" bases. Default: %(default)s')
	arg('--unique-CDR3', '--CDR3s', type=int, metavar='N', default=1,
		help='Sequences must have at least N unique CDR3s within exact sequence matches. '
		'Default: %(default)s')
	# The default for unique-J is 0 because we might work on data without
	# any assigned J genes.
	arg('--unique-J', type=int, metavar='N', default=0,
		help='Sequences must have at least N unique Js within exact sequence matches. '
		'Default: %(default)s')
	arg('--unique-D-ratio', type=float, metavar='RATIO', default=None,
		help='Discard a sequence if another allele of this gene exists '
		'such that the ratio between their Ds_exact is less than RATIO')
	arg('--unique-D-threshold', type=int, metavar='THRESHOLD', default=10,
		help='Apply the --unique-D-ratio filter only if the Ds_exact of the other '
		'allele is at least THRESHOLD')
	arg('--allow-stop', action='store_true', default=False,
		help='Allow stop codons in sequences (uses the has_stop column).'
			'Default: Do not allow stop codons.')
	arg('--whitelist', metavar='FASTA', default=[], action='append',
		help='Sequences that are never discarded or merged with others, '
			'even if criteria for discarding them would apply (except cross-mapping artifact '
			'removal, which is always performed).')
	arg('--fasta', metavar='FILE', help='Write new database in FASTA format to FILE')
	arg('tables', metavar='CANDIDATES.TAB',
		help='Tables (one or more) created by the "discover" command',
		nargs='+')


SequenceInfo = namedtuple('_SequenceInfo', ['sequence', 'name', 'clonotypes', 'exact', 'Ds_exact',
	'cluster_size', 'whitelisted', 'is_database', 'cluster_size_is_accurate', 'CDR3_start', 'row'])


class SequenceMerger(Merger):
	"""
	Merge sequences that are sufficiently similar into single entries.
	"""
	def __init__(self, max_differences, cross_mapping_ratio, clonotype_ratio, exact_ratio,
			unique_d_ratio,
			unique_d_threshold):
		super().__init__()
		self._max_differences = max_differences
		self._cross_mapping_ratio = cross_mapping_ratio
		self._clonotype_ratio = clonotype_ratio
		self._exact_ratio = exact_ratio
		self._unique_d_ratio = unique_d_ratio
		self._unique_d_threshold = unique_d_threshold

	def merged(self, s: SequenceInfo, t: SequenceInfo):
		"""
		Given two SequenceInfo objects, decide whether to discard one of them and which one.
		This is used for merging similar candidate sequences.

		Two sequences are Sequences can also be discarded if llele ratio, cross-mapping ratio or unique_d ratio criteria

		Two sequences are considered to be similar if their edit distance is at most
		max_differences (see constructor). If one of the sequences is longer, the 'overhanging'
		bases are ignored at either the 5' end or the 3' end, whichever gives the lower
		edit distance.


		If two sequences are similar and
		Sequences that are whitelisted are not discarded if the only reason for discarding
		them is their similarity
		If two sequences are similar and none of them is whitelisted, then the one with the
		higher number of unique CDR3s is kept.

		Return None if both objects should be kept.
		Return the object to keep otherwise.
		"""
		if len(s.sequence) > len(t.sequence):
			s, t = t, s  # make s always the shorter sequence

		# When computing edit distance between the two sequences, ignore the
		# bases in the 3' end that correspond to the CDR3
		s_no_cdr3 = s.sequence[:s.CDR3_start]
		t_no_cdr3 = t.sequence[:t.CDR3_start]
		if len(s_no_cdr3) != len(t_no_cdr3):
			t_prefix = t_no_cdr3[:len(s_no_cdr3)]
			t_suffix = t_no_cdr3[-len(s_no_cdr3):]
			dist_prefix = edit_distance(s_no_cdr3, t_prefix, max(self._max_differences, 1))
			dist_suffix = edit_distance(s_no_cdr3, t_suffix, max(self._max_differences, 1))
			dist = min(dist_prefix, dist_suffix)
		else:
			dist = edit_distance(s_no_cdr3, t_no_cdr3, max(self._max_differences, 1))

		# Check for cross-mapping
		# We check for dist <= 1 (and not ==1) since there may be
		# differences in the CDR3 parts of the sequences that we
		# ignore when computing edit distance.
		if self._cross_mapping_ratio and dist <= 1 and s.is_database and t.is_database:
			total_count = (s.cluster_size + t.cluster_size)
			for u, v in [(s, t), (t, s)]:
				ratio = u.cluster_size / total_count
				if u.cluster_size_is_accurate and ratio < self._cross_mapping_ratio:
					# u is probably a cross-mapping artifact of the higher-expressed v
					logger.info('%r is a cross-mapping artifact of %r (ratio %.4f)',
						u.name, v.name, ratio)
					return v

		# Check criteria based on whether the two sequences are alleles of the same gene
		if is_same_gene(s.name, t.name):
			# Check clonotype allele ratio. Somewhat similar to cross-mapping, but
			# this uses sequence names to decide whether two genes can be
			# alleles of each other and the ratio is between the
			# CDR3_clusters values
			if self._clonotype_ratio:
				for u, v in [(s, t), (t, s)]:
					ratio = u.clonotypes / v.clonotypes
					if ratio < self._clonotype_ratio:
						logger.info('Clonotype allele ratio %.4f too low for %r compared to %r',
							ratio, u.name, v.name)
						return v

			# Check exact V sequence occurrence allele ratio
			if self._exact_ratio:
				for u, v in [(s, t), (t, s)]:
					ratio = u.exact / v.exact
					if ratio < self._exact_ratio:
						logger.info('Allele ratio of exact occurrences %.4f too low for %r compared to %r',
							ratio, u.name, v.name)
						return v

			if self._unique_d_ratio:
				for u, v in [(s, t), (t, s)]:
					if u.cluster_size >= v.cluster_size and u.Ds_exact >= self._unique_d_threshold:
						ratio = v.Ds_exact / u.Ds_exact
						if ratio < self._unique_d_ratio:
							logger.info('Ds_exact ratio %.4f too low for %r compared to %r',
								ratio, v.name, u.name)
							return u

		if dist > self._max_differences:
			return None  # keep both

		# If we get here, then the two candidates are too similar. Unless they are
		# whitelisted, one needs to be discarded.
		if s.whitelisted and t.whitelisted:
			return None  # keep both
		if s.whitelisted:
			return s
		if t.whitelisted:
			return t

		# No sequence is whitelisted if we arrive here
		if s.clonotypes >= t.clonotypes:
			return s
		if len(s.sequence) < len(t.sequence):
			return t
		return t


def main(args):
	if args.unique_D_threshold <= 1:
		sys.exit('--unique-D-threshold must be at least 1')
	merger = SequenceMerger(args.max_differences, args.cross_mapping_ratio,
		clonotype_ratio=args.clonotype_ratio, exact_ratio=args.exact_ratio,
		unique_d_ratio=args.unique_D_ratio, unique_d_threshold=args.unique_D_threshold)

	whitelist = dict()
	for path in args.whitelist:
		for record in FastaReader(path):
			whitelist[record.sequence.upper()] = record.name
	logger.info('%d unique sequences in whitelist', len(whitelist))

	def whitelist_dist(sequence):
		if sequence in whitelist:
			return 0, whitelist[sequence]
		mindist = len(sequence)
		distances = []
		for seq, name in whitelist.items():
			ed = edit_distance(seq, sequence, maxdiff=mindist)
			distances.append((ed, name))
			if ed == 1:
				# We know ed does not get smaller because the
				# 'sequence in whitelist' check
				# above covers that
				return ed, name
			mindist = min(mindist, ed)
		distance, name = min(distances)
		return distance, name

	# Read in tables
	total_unfiltered = 0
	overall_table = None
	for path in args.tables:
		table = pd.read_csv(path, sep='\t')
		# TODO remove this after deprecation period
		table.rename(columns=dict(
			consensus_seqs='cluster_size',
			window_seqs='cluster_size',
			subset_seqs='cluster_size',
			CDR3_clusters='clonotypes',
		), inplace=True)

		i = list(table.columns).index('consensus')
		# whitelist_diff distinguishes between 0 and !=0 only
		# at this point. Accurate edit distances are computed later.
		whitelist_diff = [(0 if s in whitelist else -1) for s in table['consensus']]
		table.insert(i, 'whitelist_diff', pd.Series(whitelist_diff, index=table.index, dtype=int))
		table.insert(i+1, 'closest_whitelist', pd.Series('', index=table.index))

		unfiltered_length = len(table)
		table = table[table.database_diff >= args.minimum_db_diff]
		if 'N_bases' in table.columns:
			table = table[table.N_bases <= args.maximum_N]
		table = table[table.CDR3s_exact >= args.unique_CDR3]
		table = table[table.Js_exact >= args.unique_J]
		if not args.allow_stop:
			table = table[(table.has_stop == 0) | (table.whitelist_diff == 0)]
		table = table[(table.cluster_size >= args.cluster_size) | (table.whitelist_diff == 0)]
		table = table.dropna()
		logger.info('Table read from %r contains %s candidate V gene sequences. '
			'%s remain after filtering', path,
			unfiltered_length, len(table))
		if args.whitelist:
			logger.info('Of those, %d are protected by the whitelist', sum(table.whitelist_diff == 0))
		total_unfiltered += unfiltered_length

		if overall_table is None:
			overall_table = table
		else:
			overall_table.append(table)
	del table
	if len(args.tables) > 1:
		logger.info('Read %s tables with %s entries total. '
			'After filtering, %s entries remain.', len(args.tables),
			total_unfiltered, len(overall_table))

	def cluster_size_is_accurate(row):
		return bool(set(row.cluster.split(';')) & {'all', 'db'})

	for _, row in overall_table.iterrows():
		merger.add(SequenceInfo(
			sequence=row['consensus'],
			name=row['name'],
			clonotypes=row['clonotypes'],
			exact=row['exact'],
			Ds_exact=row['Ds_exact'],
			cluster_size=row['cluster_size'],
			whitelisted=row['whitelist_diff'] == 0,
			is_database=row['database_diff'] == 0,
			cluster_size_is_accurate=cluster_size_is_accurate(row),
			CDR3_start=row.get('CDR3_start', 10000),  # TODO backwards compatibility
			row=row.name,  # row.name is the index of the row. It is not row['name'].
		))

	# Discard near-duplicates
	overall_table['is_duplicate'] = pd.Series(True, index=overall_table.index, dtype=bool)
	for info in merger:
		overall_table.loc[info.row, 'is_duplicate'] = False
	overall_table = overall_table[~overall_table.is_duplicate].copy()
	del overall_table['is_duplicate']

	# Name sequences
	overall_table['name'] = overall_table['name'].apply(UniqueNamer())
	overall_table.sort_values(['name'], inplace=True)

	# Because whitelist_dist() is expensive, this is run when
	# all of the filtering has already been done
	if whitelist:
		for row in overall_table.itertuples():
			distance, name = whitelist_dist(overall_table.loc[row[0], 'consensus'])
			overall_table.loc[row[0], 'closest_whitelist'] = name
			overall_table.loc[row[0], 'whitelist_diff'] = distance
	else:
		overall_table.whitelist_diff.replace(-1, '', inplace=True)
	print(overall_table.to_csv(sep='\t', index=False, float_format='%.1f'), end='')

	if args.fasta:
		with open(args.fasta, 'w') as f:
			for _, row in overall_table.iterrows():
				print('>{}\n{}'.format(row['name'], row['consensus']), file=f)

	logger.info('%d sequences in new database', len(overall_table))



"""
Run IgDiscover

Calls Snakemake to produce all the output files.
"""
import sys
import logging
import resource
import platform
import pkg_resources
import argparse
from snakemake import snakemake
from sqt.utils import available_cpu_count
from . import __version__

from .config import Config

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--dryrun', '-n', default=False, action='store_true',
		help='Do not execute anything')
	arg('--cores', '--jobs', '-j', metavar='N', type=int, default=available_cpu_count(),
		help='Run on at most N CPU cores in parallel. '
		'Default: Use as many cores as available (%(default)s)')
	arg('--keepgoing', '-k', default=False, action='store_true',
		help='If one job fails, finish the others.')
	arg('--print-commands', '-p', default=False, action='store_true',
		help=argparse.SUPPRESS)
	arg('targets', nargs='*', default=[],
		help='File(s) to create. If omitted, the full pipeline is run.')


def main(args):
	try:
		config = Config.from_default_path()
	except FileNotFoundError as e:
		sys.exit("Pipeline configuration file {!r} not found. Please create it!".format(e.filename))

	print('IgDiscover version {} with Python {}. Configuration:'.format(__version__,
		platform.python_version()))
	for k, v in sorted(vars(config).items()):
		# TODO the following line is only necessary for non-YAML configurations
		if k.startswith('_'):
			continue
		print('   ', k, ': ', repr(v), sep='')

	if args.print_commands:
		logger.warning('The -p (--print-commands) option is ignored (commands are always printed)')
	# snakemake sets up its own logging and this cannot be easily changed
	# (setting keep_logger=True crashes), so remove our own log handler
	# for now
	logger.root.handlers = []
	snakefile_path = pkg_resources.resource_filename('igdiscover', 'Snakefile')
	success = snakemake(snakefile_path,
		snakemakepath='snakemake',  # Needed in snakemake 3.9.0
		dryrun=args.dryrun,
		cores=args.cores,
		keepgoing=args.keepgoing,
		printshellcmds=True,
		targets=args.targets if args.targets else None,
	)

	if sys.platform == 'linux' and not args.dryrun:
		cputime = resource.getrusage(resource.RUSAGE_SELF).ru_utime
		cputime += resource.getrusage(resource.RUSAGE_CHILDREN).ru_utime
		h = int(cputime // 3600)
		m = (cputime - h * 3600) / 60
		print('Total CPU time: {}h {:.2f}m'.format(h, m))

	sys.exit(0 if success else 1)



"""
Draw a dendrogram of sequences in a FASTA file.
"""
import logging
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.spatial import distance
from scipy.cluster import hierarchy
from sqt import FastaReader
from igdiscover.utils import distances


logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--mark', '--db', metavar='FASTA',
		help='Path to a FASTA file with a set of "known" sequences. Sequences '
		'in the main file that do *not* occur here will be marked with (new). '
		'If not given, no sequences will be marked (use this to compare two '
		'databases).')
	arg('--method', choices=('single', 'complete', 'weighted', 'average'),
		default='average',
		help='Linkage method. Default: "average" (=UPGMA)')
	arg('fasta', help='Path to input FASTA file')
	arg('plot', help='Path to output PDF or PNG')


class PrefixComparer:
	def __init__(self, sequences):
		self._sequences = [ s.upper() for s in sequences ]

	def __contains__(self, other):
		for seq in self._sequences:
			if seq.startswith(other) or other.startswith(seq):
				return True
		return False


def main(args):
	with FastaReader(args.fasta) as fr:
		sequences = list(fr)
	logger.info('Plotting dendrogram of %s sequences', len(sequences))
	if args.mark:
		with FastaReader(args.mark) as fr:
			mark = PrefixComparer(record.sequence for record in fr)
		labels = []
		n_new = 0
		for record in sequences:
			if record.sequence not in mark:
				extra = ' (new)'
				n_new += 1
			else:
				extra = ''
			labels.append(record.name + extra)
		logger.info('%s sequence(s) marked as "new"', n_new)
	else:
		labels = [s.name for s in sequences]
	sns.set_style("white")
	font_size = 297 / 25.4 * 72 / (len(labels) + 5)
	font_size = min(16, max(6, font_size))
	height = font_size * (len(labels) + 5) / 72
	fig = plt.figure(figsize=(210 / 25.4, height))
	matplotlib.rcParams.update({'font.size': 4})
	ax = fig.gca()
	sns.despine(ax=ax, top=True, right=True, left=True, bottom=True)
	sns.set_style('whitegrid')
	if len(sequences) >= 2:
		m = distances([s.sequence for s in sequences])
		y = distance.squareform(m)
		mindist = int(y.min())
		logger.info('Smallest distance is %s. Found between:', mindist)
		for i,j in np.argwhere(m == y.min()):
			if i < j:
				logger.info('%s and %s', labels[i], labels[j])
		l = hierarchy.linkage(y, method=args.method)
		hierarchy.dendrogram(l, labels=labels, leaf_font_size=font_size, orientation='right', color_threshold=0.95*max(l[:,2]))
	else:
		ax.text(0.5, 0.5, 'no sequences', fontsize='xx-large')
	ax.grid(False)
	fig.set_tight_layout(True)
	fig.savefig(args.plot)



"""
Discover D and J genes

The most frequent D/J sequences are considered candidates.

For J genes, candidate sequences are merged if they overlap each other.

The result table is written to standard output. Use --fasta to also
generate FASTA output.
"""
import logging
import pandas as pd
from sqt import FastaReader
from sqt.align import edit_distance
from .utils import Merger, merge_overlapping, unique_name, is_same_gene, slice_arg
from .table import read_table, fix_columns

logger = logging.getLogger(__name__)

MINIMUM_CANDIDATE_LENGTH = 5


def add_arguments(parser):
	arg = parser.add_argument
	arg('--database', metavar='FASTA',
		help='FASTA file with reference gene sequences')
	arg('--merge', default=None, action='store_true', help='Merge overlapping genes. '
		'Default: Enabled for D, disabled for J and V.')
	arg('--no-merge', dest='merge', action='store_false', help='Do not merge overlapping genes')
	arg('--gene', default='J', choices=('V', 'D', 'J'),
		help='Which gene category to discover. Default: %(default)s')
	arg('--allele-ratio', type=float, metavar='RATIO', default=0.1,
		help='Required allele ratio. Works only for genes named "NAME*ALLELE". Default: %(default)s')
	arg('--cross-mapping-ratio', type=float, metavar='RATIO', default=None,
		help='Ratio for detection of cross-mapping artifacts. Default: %(default)s')
	arg('--min-count', metavar='N', type=int, default=None,
		help='Omit candidates with fewer than N exact occurrences in the input table. '
			'Default: 10 for D, 100 for V and J')

	# --gene=D options
	arg('--d-core-length', metavar='L', type=int, default=6,
		help='Use only D core regions that have at least length L (only '
			'applies when --gene=D). Default: %(default)s')
	arg('--d-core', type=slice_arg, default=slice(2, -2),
		help='D core region location (only applies when --gene=D). '
			'Default: %(default)s')

	arg('--fasta', help='Write discovered sequences to FASTA file')
	arg('table', help='Table with parsed and filtered IgBLAST results')


class Candidate:
	__slots__ = ('name', 'sequence', 'exact_occ', 'max_count', 'other_genes', 'db_name',
		'db_distance', 'cdr3s', 'missing')

	def __init__(self, name, sequence, exact_occ=0, max_count=0, cdr3s=None, other_genes=None,
			db_name=None, db_distance=None):
		self.name = name
		self.sequence = sequence
		self.exact_occ = exact_occ
		self.max_count = max_count  # an upper bound for the exact_occ
		self.cdr3s = cdr3s if cdr3s is not None else set()
		self.other_genes = other_genes if other_genes is not None else set()
		self.db_name = db_name
		self.db_distance = db_distance
		self.missing = ''

	@property
	def unique_CDR3(self):
		return len(self.cdr3s)

	def __repr__(self):
		return 'Candidate({sequence!r}, exact_occ={exact_occ}, max_count={max_count}, ...)'.format(
			sequence=self.sequence,
			exact_occ=self.exact_occ,
			max_count=self.max_count,
		)


class OverlappingSequenceMerger(Merger):
	"""
	Merge sequences that overlap
	"""
	def merged(self, s, t):
		"""
		Merge two sequences if they overlap. If they should not be merged,
		None is returned.
		"""
		m = merge_overlapping(s.sequence, t.sequence)
		if m is not None:
			return Candidate(s.name, m, max_count=t.max_count + s.max_count)

		return None


class AlleleRatioMerger(Merger):
	"""
	Discard sequences with too low allele ratio
	"""
	def __init__(self, allele_ratio, cross_mapping_ratio):
		super().__init__()
		self._allele_ratio = allele_ratio
		self._cross_mapping_ratio = cross_mapping_ratio

	def merged(self, s, t):
		"""
		Merge two sequences if they overlap. If they should not be merged,
		None is returned.
		"""
		# TODO copy-and-pasted from germlinefilter
		#
		# Check allele ratio. Somewhat similar to cross-mapping, but
		# this uses sequence names to decide whether two genes can be
		# alleles of each other and the ratio is between the CDR3s_exact
		# values
		if self._allele_ratio and is_same_gene(s.name, t.name):
			for u, v in [(s, t), (t, s)]:
				if v.unique_CDR3 == 0:
					continue
				ratio = u.unique_CDR3 / v.unique_CDR3
				if ratio < self._allele_ratio:
					# logger.info('Allele ratio %.4f too low for %r compared to %r',
					# 	ratio, u.name, v.name)
					return v

		if self._cross_mapping_ratio:
			# When checking for cross mapping, ignore overhanging bases in the 5' end.
			# Example:
			# ---ACTACGACTA...
			# XXX|||||X||||
			# ATTACTACTACTA...
			if len(t.sequence) < len(s.sequence):
				t, s = s, t  # s is now the shorter sequence
			t_seq = t.sequence[len(t.sequence) - len(s.sequence):]
			s_seq = s.sequence
			dist = edit_distance(s_seq, t_seq, 1)
			if dist > 1:
				return None
			total_occ = (s.exact_occ + t.exact_occ)
			if total_occ == 0:
				return None
			for u, v in [(s, t), (t, s)]:
				ratio = u.exact_occ / total_occ
				if ratio < self._cross_mapping_ratio:
					# u is probably a cross-mapping artifact of the higher-expressed v
					logger.info('%r is a cross-mapping artifact of %r (ratio %.4f)',
						u.name, v.name, ratio)
					return v

		return None


def count_occurrences(candidates, table_path, search_columns, other_gene, other_errors, merge):
	"""
	Count how often each candidate sequence occurs in the input table.
	The columns named in search_columns are concatenated and searched.
	This circumvents inaccurate IgBLAST alignment boundaries.
	Rows where the column named by other_errors is not zero are ignored.

	The input table is read in chunks to reduce memory usage.

	candidates -- list of candidates
	table_path -- path to input table
	search_columns -- which columns to search. The contained strings are
	    concatenated and then searched.
	merge -- If True, stop searching for other candidates in a single row
	    after one candidate has been found.

	The following attributes of the candidates are updated:

	- exact_occ
	- other_genes
	- cdr3s

	Return the updated list of candidates.
	"""
	candidates_map = {c.sequence: c for c in candidates}

	# Speed up search by looking for the most common sequences first
	search_order = sorted(candidates_map, key=lambda s: candidates_map[s].max_count, reverse=True)
	cols = [other_gene, 'V_errors', 'J_errors', 'CDR3_nt'] + search_columns
	for chunk in pd.read_csv(table_path, usecols=cols, chunksize=10000, sep='\t'):
		fix_columns(chunk)
		chunk = chunk[chunk[other_errors] == 0]
		# concatenate search columns
		if len(chunk) == 0:  # TODO that this is needed is possibly a pandas bug
			continue
		chunk['haystack'] = chunk.loc[:, search_columns].astype(str).sum(axis=1)
		chunk['haystack'] = chunk['haystack'].str.replace('(', '').replace(')', '')

		for row in chunk.itertuples():
			for needle in search_order:
				if needle in row.haystack:
					candidate = candidates_map[needle]
					candidate.exact_occ += 1  # TODO += row.count?
					candidate.other_genes.add(getattr(row, other_gene))
					candidate.cdr3s.add(row.CDR3_nt)
					if merge:
						# When overlapping candidates have been merged,
						# there will be no other pattern that is a
						# substring of the current search pattern.
						break
	return candidates_map.values()


def discard_substring_occurrences(candidates):
	"""
	Filter a candidate list by discarding candidates whose sequences are
	substrings of another candidate’s sequence
	"""
	# Shorter sequences first
	candidates = sorted(candidates, key=lambda c: len(c.sequence))
	for i, short in enumerate(candidates):
		for long in candidates[i+1:]:
			if short.sequence in long.sequence:
				break
		else:
			# no substring occurrence - keep this candidate
			yield short


def sequence_candidates(table, column, minimum_length, core=slice(None, None)):
	"""
	Generate candidates by clustering all sequences in a column
	(usually V_nt, J_nt). At least two occurrences are required.

	core -- a slice object. If given, the strings in the column are
	       sliced before being clustered.
	"""
	for sequence, occ in table[column].str[core].value_counts().items():
		if len(sequence) >= minimum_length and occ >= 2:
			yield Candidate(None, sequence, max_count=occ)


def print_table(candidates, other_gene, missing):
	columns = ['name', 'exact_occ', other_gene + 's', 'CDR3s', 'database', 'database_diff', 'sequence']
	if missing:
		columns.append('missing')
	print(*columns, sep='\t')
	for candidate in candidates:
		columns = [
			candidate.name,
			candidate.exact_occ,
			len(set(candidate.other_genes)),
			candidate.unique_CDR3,
			candidate.db_name if candidate.db_name is not None else '',
			candidate.db_distance if candidate.db_distance is not None else -1,
			candidate.sequence
		]
		if missing:
			columns.append(candidate.missing)
		print(*columns, sep='\t')


def main(args):
	if args.database:
		with FastaReader(args.database) as fr:
			database = list(fr)
		logger.info('Read %d sequences from %r', len(database), args.database)
	else:
		database = None
	column = {'V': 'V_nt', 'J': 'J_nt', 'D': 'D_region'}[args.gene]
	other = 'V' if args.gene in ('D', 'J') else 'J'
	other_gene = other + '_gene'
	other_errors = other + '_errors'
	table = read_table(args.table,
		usecols=['count', 'V_gene', 'J_gene', 'V_errors', 'J_errors', column, 'CDR3_nt'])
	logger.info('Table with %s rows read', len(table))
	table = table[table[other_errors] == 0]
	logger.info('Keeping %s rows that have no %s mismatches', len(table), other)

	if args.merge is None:
		args.merge = args.gene == 'D'
	if args.min_count is None:
		args.min_count = 10 if args.gene == 'D' else 100

	if args.gene == 'D':
		candidates = sequence_candidates(
			table, column, minimum_length=args.d_core_length, core=args.d_core)
	else:
		candidates = sequence_candidates(
			table, column, minimum_length=MINIMUM_CANDIDATE_LENGTH)

	candidates = list(candidates)
	logger.info('Collected %s unique %s sequences', len(candidates), args.gene)
	candidates = list(discard_substring_occurrences(candidates))
	logger.info('Removing candidate sequences that occur within others results in %s candidates',
		len(candidates))

	if args.merge:
		logger.info('Merging overlapping sequences ...')
		# Merge candidate sequences that overlap. If one candidate is longer than
		# another, this is typically a sign that IgBLAST has not extended the
		# alignment long enough.
		merger = OverlappingSequenceMerger()
		for candidate in candidates:
			merger.add(candidate)
		logger.info('After merging overlapping %s sequences, %s remain', args.gene, len(merger))
		candidates = list(merger)
	del table

	logger.info('Counting occurrences ...')
	if args.gene == 'D':
		search_columns = ['VD_junction', 'D_region', 'DJ_junction']
	else:
		search_columns = ['genomic_sequence']
	records = count_occurrences(candidates, args.table, search_columns, other_gene, other_errors,
		args.merge)

	logger.info('%d records', len(records))
	# Assign names etc.
	if database:
		for record in records:
			distances = [(edit_distance(db.sequence, record.sequence), db) for db in database]
			record.db_distance, closest = min(distances, key=lambda x: x[0])
			record.db_name = closest.name

			if record.db_distance == 0:
				record.name = closest.name
			else:
				# Exact db sequence not found, is there one that contains
				# this candidate as a substring?
				is_substring = [db for db in database
					if db.sequence.find(record.sequence) != -1]
				if is_substring:
					db = is_substring[-1]
					start = db.sequence.find(record.sequence)
					prefix = db.sequence[:start]
					suffix = db.sequence[start + len(record.sequence):]
					record.name = db.name
					record.missing = '{}...{}'.format(prefix, suffix)
				else:
					record.name = unique_name(closest.name, record.sequence)
	else:
		for record in records:
			record.name = unique_name(args.gene, record.sequence)

	# Filter by allele ratio
	if args.allele_ratio or args.cross_mapping_ratio:
		arm = AlleleRatioMerger(args.allele_ratio, args.cross_mapping_ratio)
		arm.extend(records)
		records = list(arm)
		logger.info('After filtering by allele ratio and/or cross-mapping ratio, %d candidates remain',
			len(records))

	records = sorted(records, key=lambda r: (r.exact_occ, r.sequence), reverse=True)
	records = [r for r in records if r.exact_occ >= args.min_count]
	print_table(records, other_gene, missing=args.gene == 'D')

	if args.fasta:
		with open(args.fasta, 'w') as f:
			for record in sorted(records, key=lambda r: r.name):
				print('>{}\n{}'.format(record.name, record.sequence), file=f)

	logger.info('Wrote %d genes', len(records))



import sys
from setuptools import setup
import versioneer


if sys.version_info < (3, 4):
	sys.stdout.write("At least Python 3.4 is required.\n")
	sys.exit(1)

with open('README.rst', encoding='utf-8') as f:
	long_description = f.read()

setup(
	name = 'igdiscover',
	version = versioneer.get_version(),
	cmdclass = versioneer.get_cmdclass(),
	author = 'Marcel Martin',
	author_email = 'marcel.martin@scilifelab.se',
	url = 'https://igdiscover.se/',
	description = 'Analyze antibody repertoires and discover new V genes',
	long_description = long_description,
	license = 'MIT',
	entry_points = {'console_scripts': ['igdiscover = igdiscover.__main__:main']},
	packages = ['igdiscover'],
	package_data = {'igdiscover': ['igdiscover.yaml', 'Snakefile', 'empty.aux']},
	install_requires = [
		'sqt>=0.8.0',
		'pandas>=0.16.2',
		'numpy',
		'matplotlib>=1.5.0',
		'seaborn==0.7.1',
		'snakemake>=3.9.0',
		'cutadapt',
		'scipy>=0.16.1',
		'xopen>=0.3.2',
		'ruamel.yaml',
	],
	classifiers = [
		"Development Status :: 3 - Alpha",
		"Environment :: Console",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: MIT License",
		"Natural Language :: English",
		"Programming Language :: Python :: 3",
		"Topic :: Scientific/Engineering :: Bio-Informatics"
	]
)



"""
Discover candidate new V genes within a single antibody library.

Existing V sequences are grouped by their V gene assignment, and within each
group, consensus sequences are computed.
"""
import csv
import hashlib
import logging
import sys
from collections import namedtuple, Counter
import multiprocessing
import random
import itertools
import numpy as np
import pandas as pd
from sqt import SequenceReader
from sqt.align import edit_distance
from sqt.utils import available_cpu_count

from .table import read_table
from .utils import iterative_consensus, unique_name, downsampled, SerialPool, Merger, has_stop
from .cluster import cluster_sequences, single_linkage

logger = logging.getLogger(__name__)


MINGROUPSIZE = 5
MAXIMUM_SUBSAMPLE_SIZE = 1600


def add_arguments(parser):
	arg = parser.add_argument
	arg('--threads', '-j', type=int, default=min(4, available_cpu_count()),
		help='Number of threads. Default: no. of available CPUs, but at most 4')
	arg('--seed', type=int, default=None,
		help='Seed value for random numbers for reproducible runs.')
	arg('--consensus-threshold', '-t', metavar='PERCENT', type=float, default=60,
		help='Threshold for consensus computation. Default: %(default)s%%')
	arg('--gene', '-g', action='append', default=[],
		help='Compute consensus for this gene. Can be given multiple times. '
			'Default: Compute for all genes.')
	arg('--left', '-l', type=float, metavar='ERROR-RATE',
		help='For consensus, include only sequences that have at least this '
			'error rate (in percent). Default: %(default)s', default=0)
	arg('--right', '-r', type=float, metavar='ERROR-RATE',
		help='For consensus, include only sequences that have at most this '
			'error rate (in percent). Default: %(default)s', default=100)
	arg('--window-width', '-w', type=float, metavar='PERCENT',
		help='Compute consensus for all PERCENT-wide windows. Set to 0 to '
			'disable. Default: %(default)s', default=2)
	arg('--no-cluster', dest='cluster', action='store_false', default=True,
		help='Do not run linkage cluster analysis.')
	arg('--cluster-exact', metavar='N', type=int, default=0,
		help='Treat N exact occurrences of a sequence as a cluster. '
			'Default: Do not cluster exact occurrences')
	arg('--max-n-bases', type=int, default=0, metavar='MAXN',
		help='Remove rows that have more than MAXN "N" nucleotides. If >0, an '
			'N_bases column is added. Default: %(default)s')
	arg('--subsample', metavar='N', type=int, default=1000,
		help='When clustering, use N randomly chosen sequences. Default: %(default)s')
	arg('--table-output', '-o', metavar='DIRECTORY',
		help='Output tables for all analyzed genes to DIRECTORY. '
			'Files will be named <GENE>.tab.')
	arg('--database', metavar='FASTA', default=None,
		help='FASTA file with V genes. If provided, differences between consensus '
			'and database will be computed.')
	arg('--ignore-J', action='store_true', default=False,
		help='Include also rows without J assignment or J%%SHM>0.')
	arg('--exact-copies', metavar='N', type=int, default=1,
		help='When subsampling, first pick rows whose V gene sequences'
			'have at least N exact copies in the input. Default: %(default)s')
	arg('--d-evalue', metavar='EVALUE', type=float, default=1E-4,
		help='For Ds_exact, require D matches with an E-value of '
			'at most EVALUE. Default: %(default)s')
	arg('--d-coverage', '--D-coverage', metavar='COVERAGE', type=float, default=70,
		help='For Ds_exact, require D matches with a minimum D '
			'coverage of COVERAGE (in percent). Default: %(default)s)')
	arg('--clonotype-diff', metavar='DIFFERENCES', type=int, default=6,
		help='When clustering CDR3s to computer the no. of clonotypes, allow DIFFERENCES '
		     'between (nucleotide-)sequences. Default: %(default)s')
	arg('table', help='Table with parsed IgBLAST results')


Groupinfo = namedtuple('Groupinfo',
	'count unique_D unique_J unique_CDR3 shared_CDR3_ratio clonotypes unique_barcodes')

SiblingInfo = namedtuple('SiblingInfo', 'sequence requested name group')


class SiblingMerger(Merger):
	"""
	Merge very similar consensus sequences into single entries. This could be
	seen as a type of clustering using very specific criteria. Two sequences
	are merged if one is the prefix of the other, allowing differences where
	one of the sequences has an 'N' base.
	"""
	def merged(self, s, t):
		chars = []
		for c1, c2 in itertools.zip_longest(s.sequence, t.sequence):
			if c1 is None:
				c = c2
			elif c2 is None:
				c = c1
			elif c1 == 'N':
				c = c2
			elif c2 == 'N':
				c = c1
			elif c1 != c2:
				return None
			else:
				assert c1 == c2
				c = c1
			chars.append(c)
		seq = ''.join(chars)
		requested = s.requested or t.requested
		name = s.name + ';' + t.name
		# take union of groups
		group = pd.concat([s.group, t.group]).groupby(level=0).last()
		return SiblingInfo(seq, requested, name, group)


class Discoverer:
	"""
	Discover candidates for novel V genes.
	"""
	def __init__(self, database, windows, left, right, cluster, cluster_exact,
			table_output, consensus_threshold, downsample,
			clonotype_differences, cluster_subsample_size,
			max_n_bases, exact_copies, d_coverage, d_evalue, seed, cdr3_counts):
		self.database = database
		self.windows = windows
		self.left = left
		self.right = right
		self.cluster = cluster
		self.cluster_exact = cluster_exact
		self.table_output = table_output
		self.consensus_threshold = consensus_threshold
		self.downsample = downsample
		self.clonotype_differences = clonotype_differences
		self.cluster_subsample_size = cluster_subsample_size
		self.max_n_bases = max_n_bases
		self.exact_copies = exact_copies
		self.d_coverage = d_coverage
		self.d_evalue = d_evalue
		self.seed = seed
		self.cdr3_counts = cdr3_counts

	def _sibling_sequence(self, gene, group):
		"""
		For a given group, compute a consensus sequence over the V gene sequences
		in that group.

		If the found sibling is slightly longer or shorter than the version in
		the database, adjust it so it corresponds to the database version exactly.
		"""
		sequence = iterative_consensus(list(group.V_nt), program='muscle-medium',
			threshold=self.consensus_threshold/100,
			maximum_subsample_size=self.downsample)
		if gene in self.database:
			database_sequence = self.database[gene]
			if sequence.startswith(database_sequence) or database_sequence.startswith(sequence):
				return database_sequence
		return sequence

	@staticmethod
	def _guess_chain(group):
		"""
		Return a guess for the chain type of a given group
		"""
		return Counter(group.chain).most_common()[0][0]

	@staticmethod
	def _guess_cdr3_start(group):
		"""
		Return a guess for the CDR3 start within sequences in the given group
		"""
		return Counter(group.V_CDR3_start).most_common()[0][0]

	def count_unique_d(self, table):
		g = table[(table.D_errors == 0) &
			(table.D_covered >= self.d_coverage) &
			(table.D_evalue <= self.d_evalue)]
		return len(set(s for s in g.D_gene if s))

	@staticmethod
	def count_unique_barcodes(group):
		return len(set(s for s in group.barcode if s))

	def count_clonotypes(self, table):
		"""
		Cluster sequences by edit distance and return the number of clusters.

		The sequences are first group by their J assignment and cluster
		numbers are computed within these groups separately, then summed up.
		"""
		distance = self.clonotype_differences

		def linked(s, t):
			return edit_distance(s, t, distance) <= distance

		total = 0
		for j_gene, group in table.groupby('J_gene'):
			sequences = list(set(s for s in group.CDR3_nt if s))
			components = single_linkage(sequences, linked)
			total += len(components)
		return total

	def _cluster_siblings(self, gene, table):
		"""Find candidates by clustering sequences assigned to one gene"""

		if self.exact_copies > 1:
			# Preferentially pick those sequences (for subsampling) that have
			# multiple exact copies, then fill up with the others
			exact_group = table[table.copies >= self.exact_copies]
			indices = downsampled(list(exact_group.index), self.cluster_subsample_size)
			if len(indices) < self.cluster_subsample_size:
				not_exact_group = table[table.copies < self.exact_copies]
				indices.extend(downsampled(list(not_exact_group.index),
					self.cluster_subsample_size - len(indices)))
		else:
			indices = downsampled(list(table.index), self.cluster_subsample_size)
		# Ignore CDR3 part of the V sequence for clustering
		sequences_no_cdr3 = list(table.V_no_CDR3.loc[indices])
		df, linkage, clusters = cluster_sequences(sequences_no_cdr3, MINGROUPSIZE)
		n_clusters = len(set(clusters))
		logger.info('%6d %s assignments generated %d cluster%s',
			len(table), gene, n_clusters, 's' if n_clusters != 1 else '')
		cluster_indices = [[] for _ in range(max(clusters) + 1)]
		for i, cluster_id in enumerate(clusters):
			cluster_indices[cluster_id].append(indices[i])

		cl = 0
		for ind in cluster_indices:
			group = table.loc[ind]
			if len(group) < MINGROUPSIZE:
				continue
			sibling = self._sibling_sequence(gene, group)
			cl += 1
			yield SiblingInfo(
				sequence=sibling, requested=False, name='cl{}'.format(cl), group=group)

	def _window_siblings(self, gene, table):
		"""
		Find candidates by clustering sequences that have a similar number of differences
		to the reference sequence
		"""
		for left, right in self.windows:
			left, right = float(left), float(right)
			group = table[(left <= table.V_SHM) & (table.V_SHM < right)]
			if len(group) < MINGROUPSIZE:
				continue
			sibling = self._sibling_sequence(gene, group)
			if left == int(left):
				left = int(left)
			if right == int(right):
				right = int(right)
			requested = (left, right) == (self.left, self.right)
			if (left, right) == (0, 100):
				name = 'all'
			else:
				name = '{}-{}'.format(left, right)
			yield SiblingInfo(sequence=sibling, requested=requested, name=name, group=group)

	def _cluster_exact_candidates(self, gene, table):
		index = 1
		for sequence, group in table.groupby('V_nt'):
			if len(group) >= self.cluster_exact:
				name = 'ex{}'.format(index)
				index += 1
				yield SiblingInfo(sequence, False, name, group)

	def _collect_siblings(self, gene, group):
		"""
		gene -- gene name
		group -- pandas.DataFrame of sequences assigned to that gene

		Yield SiblingInfo objects
		"""
		group = group.copy()
		# the original reference sequence for all the IgBLAST assignments in this group
		database_sequence = self.database.get(gene, None)
		database_sequence_found = False

		candidate_iterators = [self._window_siblings(gene, group)]
		if self.cluster:
			candidate_iterators.append(self._cluster_siblings(gene, group))
		if self.cluster_exact:
			candidate_iterators.append(self._cluster_exact_candidates(gene, group))
		for sibling in itertools.chain(*candidate_iterators):
			if sibling.sequence == database_sequence:
				database_sequence_found = True
			yield sibling

		if database_sequence:
			# If this is a database sequence and there are some exact occurrences of it,
			# add it to the list of candidates even if it has not been found as a
			# cluster.
			group_in_window = group[group.V_errors == 0]
			if len(group_in_window) >= MINGROUPSIZE:
				if not database_sequence_found:
					logger.info('Database sequence %r seems to be expressed, but is missing from '
						'candidates. Re-adding it.', gene)
				yield SiblingInfo(database_sequence, False, 'db', group_in_window)

	def set_random_seed(self, name):
		"""Set random seed depending on gene name and seed given to constructor"""
		h = hashlib.md5(name.encode()).digest()[:4]
		n = int.from_bytes(h, byteorder='big')
		random.seed(n + self.seed)

	def __call__(self, args):
		"""
		Discover new V genes. args is a tuple (gene, group)
		gene -- name of the gene
		group -- a pandas DataFrame with the group corresponding to the gene
		"""
		gene, group = args
		self.set_random_seed(gene)
		siblings = SiblingMerger()
		for sibling in self._collect_siblings(gene, group):
			siblings.add(sibling)

		candidates = []
		for sibling_info in siblings:
			sibling = sibling_info.sequence
			n_bases = sibling.count('N')
			if n_bases > self.max_n_bases:
				logger.debug('Sibling %s has too many N bases', sibling_info.name)
				continue

			sibling_no_cdr3 = sibling[:self._guess_cdr3_start(group)]
			group_exact_v = group[group.V_no_CDR3 == sibling_no_cdr3]
			groups = (
				('window', sibling_info.group),
				('exact', group_exact_v))
			del sibling_no_cdr3

			other_cdr3_counts = self.cdr3_counts - Counter(s for s in sibling_info.group.CDR3_nt if s)
			info = dict()
			for key, g in groups:
				cdr3_counts = Counter(s for s in g.CDR3_nt if s)
				unique_cdr3 = len(cdr3_counts)
				shared_cdr3_ratio = len(other_cdr3_counts & cdr3_counts) / unique_cdr3 if unique_cdr3 > 0 else 0
				unique_j = len(set(s for s in g.J_gene if s))
				clonotypes = self.count_clonotypes(g)
				unique_d = self.count_unique_d(g)
				unique_barcodes = self.count_unique_barcodes(g)
				count = len(g.index)
				info[key] = Groupinfo(count=count, unique_D=unique_d, unique_J=unique_j,
					unique_CDR3=unique_cdr3, shared_CDR3_ratio=shared_cdr3_ratio,
					clonotypes=clonotypes,
					unique_barcodes=unique_barcodes)
			if gene in self.database:
				database_diff = edit_distance(sibling, self.database[gene])
			else:
				database_diff = None

			# Build the Candidate
			sequence_id = gene if database_diff == 0 else unique_name(gene, sibling)
			chain = self._guess_chain(sibling_info.group)
			cdr3_start = self._guess_cdr3_start(sibling_info.group)
			try:
				ratio = info['exact'].count / info['exact'].unique_CDR3
			except ZeroDivisionError:
				ratio = 0
			candidate = Candidate(
				name=sequence_id,
				source=gene,
				chain=chain,
				cluster=sibling_info.name,
				cluster_size=info['window'].count,
				Js=info['window'].unique_J,
				CDR3s=info['window'].unique_CDR3,
				exact=info['exact'].count,
				barcodes_exact=info['exact'].unique_barcodes,
				Ds_exact=info['exact'].unique_D,
				Js_exact=info['exact'].unique_J,
				CDR3s_exact=info['exact'].unique_CDR3,
				clonotypes=info['exact'].clonotypes,
				CDR3_exact_ratio=ratio,
				CDR3_shared_ratio=info['exact'].shared_CDR3_ratio,
				N_bases=n_bases,
				database_diff=database_diff,
				has_stop=has_stop(sibling),
				CDR3_start=cdr3_start,
				consensus=sibling,
			)
			candidates.append(candidate)
		return candidates


class Candidate(namedtuple('_Candidate', [
	'name',
	'source',
	'chain',
	'cluster',
	'cluster_size',
	'Js',
	'CDR3s',
	'exact',
	'barcodes_exact',
	'Ds_exact',
	'Js_exact',
	'CDR3s_exact',
	'clonotypes',
	'CDR3_exact_ratio',
	'CDR3_shared_ratio',
	'N_bases',
	'database_diff',
	'has_stop',
	'CDR3_start',
	'consensus',
])):
	def formatted_dict(self):
		d = self._asdict()
		d['has_stop'] = int(d['has_stop'])
		for name in 'CDR3_exact_ratio', 'CDR3_shared_ratio':
			d[name] = '{:.2f}'.format(d[name])
		return d


def count_prefixes(sequences):
	"""
	Count how often each sequence occurs in the given list of
	sequences. If one sequence is the prefix of another one,
	they are considered to be 'identical'.

	Return a dictionary that maps sequence to count.

	>>> r = count_prefixes(['A', 'BB', 'CD', 'CDE', 'CDEF'])
	>>> r == {'A': 1, 'BB': 1, 'CD': 3, 'CDE': 3, 'CDEF': 3}
	True
	"""
	sequences = sorted(sequences)
	sequences.append('GUARD')
	prev = 'X'
	start = 0
	count = dict()
	for i, s in enumerate(sequences):
		if not s.startswith(prev):
			# end of a run
			for j in range(start, i):
				count[sequences[j]] = i - start
			start = i
		prev = s
	return count


def main(args):
	if args.database:
		with SequenceReader(args.database) as sr:
			database = {record.name: record.sequence.upper() for record in sr}
	else:
		database = dict()

	if args.seed:
		seed = args.seed
	else:
		seed = random.randrange(10**6)
		logger.info('Use --seed=%d to reproduce this run', seed)

	table = read_table(args.table, usecols=('name', 'chain', 'V_gene', 'D_gene', 'J_gene', 'V_nt',
		'CDR3_nt', 'barcode', 'V_CDR3_start', 'V_SHM', 'J_SHM', 'D_covered', 'D_evalue', 'V_errors',
		'D_errors', 'J_errors'))
	table['V_no_CDR3'] = [s[:start] if start != 0 else s for s, start in
		zip(table.V_nt, table.V_CDR3_start)]

	logger.info('%s rows read', len(table))
	if not args.ignore_J:
		# Discard rows with any mutation within J at all
		table = table[table.J_SHM == 0][:]
		logger.info('%s rows remain after discarding J%%SHM > 0', len(table))

	if args.exact_copies > 1:
		multiplicities = count_prefixes(table.V_no_CDR3)
		table['copies'] = table.V_no_CDR3.map(multiplicities)
		logger.info('%s rows contain V sequences with at least %s copies',
			sum(table.copies >= args.exact_copies), args.exact_copies)

	columns = list(Candidate._fields)
	if not args.max_n_bases:
		columns.remove('N_bases')
	writer = csv.DictWriter(sys.stdout, fieldnames=columns, delimiter='\t', extrasaction='ignore')
	writer.writeheader()
	genes = set(args.gene)
	if args.window_width:
		windows = [(start, start + args.window_width)
			for start in np.arange(0, 20, args.window_width)]
		logger.info('Using an error rate window of %.1f%% to %.1f%%', args.left, args.right)
		windows.append((args.left, args.right))
	else:
		windows = []

	groups = []
	for gene, group in table.groupby('V_gene'):
		if genes and gene not in genes:
			continue
		if len(group) < MINGROUPSIZE:
			continue
		groups.append((gene, group))

	cdr3_counts = Counter(s for s in table.CDR3_nt if s)
	logger.info('%s unique CDR3s detected overall', len(cdr3_counts))
	discoverer = Discoverer(database, windows, args.left, args.right, args.cluster,
		args.cluster_exact, args.table_output, args.consensus_threshold,
		MAXIMUM_SUBSAMPLE_SIZE, clonotype_differences=args.clonotype_diff,
		cluster_subsample_size=args.subsample,
		max_n_bases=args.max_n_bases, exact_copies=args.exact_copies,
		d_coverage=args.d_coverage, d_evalue=args.d_evalue,
		seed=seed, cdr3_counts=cdr3_counts)
	n_consensus = 0

	Pool = SerialPool if args.threads == 1 else multiprocessing.Pool
	with Pool(args.threads) as pool:
		for candidates in pool.imap(discoverer, groups, chunksize=1):
			for candidate in candidates:
				writer.writerow(candidate.formatted_dict())
			sys.stdout.flush()
			n_consensus += len(candidates)
	logger.info('%s consensus sequences for %s gene(s) computed', n_consensus, len(groups))
